#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["CPW"]

from .constants import *
import jax.numpy as np
from .helpers import kokp



class CPW:
    def __init__(self, L, w, s, h, thickness=None, epsilon=1, grounded=False):
        self.L = L * 1e-3
        self.w = w * 1e-3
        self.h = h * 1e-3
        self.s = s * 1e-3
        self.thickness = thickness if thickness is None else thickness * 1e-3
        self.epsilon = epsilon
        self.grounded = grounded

        if grounded and thickness is None:
            raise ValueError("thinkness must be finite when grounded is True")
        self.k1 = w / (w + 2 * s)
        self.k2 = (
            0
            if self.thickness is None
            else np.sinh(pi * self.w / (4 * self.thickness))
            / np.sinh(pi * (self.w + 2 * self.s) / (4 * self.thickness))
        )
        self.k3 = (
            0
            if self.thickness is None
            else np.tanh(pi * self.w / (4 * self.thickness))
            / np.tanh(pi * (self.w + 2 * self.s) / (4 * self.thickness))
        )

    @property
    def effective_epsilon(self):
        if self.thickness is None:
            return (self.epsilon + 1) / 2
        else:
            if self.grounded:
                q = kokp(self.k3) / (kokp(self.k1) + kokp(self.k3))
                return 1 + q * (self.epsilon - 1)
            else:
                return 1 + (self.epsilon - 1) / 2 * kokp(self.k2) / kokp(self.k1)

    @property
    def impedance(self):
        if self.grounded:
            return (
                60
                * pi
                / (self.effective_epsilon) ** 0.5
                / (kokp(self.k1) + kokp(self.k3))
            )
        else:
            return 30 * pi / (self.effective_epsilon) ** 0.5 / kokp(self.k1)



    def get_S(self,frequency):
        omega = frequency * 2 * pi * 1e9
        k0 = omega / c
        Z0 = 50
        # Zc = 3
        Zc = self.impedance
        # Zc = Z0 * np.real(1 + S11_target) / (1 - S11_target)
        gamma = 1j * (self.effective_epsilon) ** 0.5 * k0
        sh = np.sinh(gamma * self.L)
        ch = np.cosh(gamma * self.L)
        denom = 2 * Zc * Z0 * ch + (Zc ** 2 + Z0 ** 2) * sh
        S11 = (Zc ** 2 - Z0 ** 2) * sh / denom
        S12 = 2 * Zc * Z0 / denom
        S21 = S12
        S22 = S11
        return dict(S11=S11, S12=S12, S21=S21, S22=S22)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT



import jax.numpy as np
from .constants import *



deg2rad = pi / 180
rad2deg = 1 / deg2rad

def complementary(k):
    return (1 - k ** 2) ** 0.5


def kokp(k):
    kp = complementary(k)
    val0 = pi / np.log(2 * (1 + kp ** 0.5) / (1 - kp ** 0.5))
    val1 = np.log(2 * (1 + k ** 0.5) / (1 - k ** 0.5)) / pi
    return np.where(np.abs(k) < 1 / 2 ** 0.5, val0, val1)


def unwrap(phi):
    return np.unwrap(phi * deg2rad) * rad2deg


def meas2complex(A, phi):
    return 10 ** (A / 20) * np.exp(1j * phi * deg2rad)


def complex2aphi(z):
    return np.abs(z), np.unwrap(np.angle(z)) * 180 / pi



# 
# from scipy.special import ellipk
# 
# def K(k):
#     return ellipk(k)
# 
# def Kp(k):
#     return K(complementary(k))
# 
# def kokp_scipy(k):
#     return K(k)/Kp(k)
# 
# 
# k = np.linspace(0.0,1,100)
# 
# plt.figure()
# plt.plot(k,kokp(k))
# plt.plot(k,kokp_scipy(k),"o")
# xsx



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

__all__ = ["Extractor"]

import jax.numpy as np
from jax import grad, jit
from jax.config import config
import numpy as npo
import nlopt
from .constants import *


def mse(a, b):
    return np.mean(np.abs(a - b) ** 2)


def eps2opt(eps):
    return npo.hstack([eps.real, eps.imag])


class Extractor:
    def __init__(
        self,
        frequencies,
        Smeas,
        guess,
        cpw,
        options=None,
        eps_min=-1e9 - 1e9 * 1j,
        eps_max=1e9 + 1e9 * 1j,
        compile=True,
        verbose=True,
        backend="nlopt",
        bounds=None,
    ):

        if backend not in ["scipy", "nlopt"]:
            raise ValueError("backend must be scipy or nlopt")
        self.bounds = bounds
        self.backend = backend
        self.cpw = cpw
        self.frequencies = frequencies
        self.Smeas = Smeas
        self.guess = guess
        self.nfreq = len(frequencies)
        self.nvar = 2 * self.nfreq
        self.verbose = verbose
        self.compile = compile
        default_options = {
            "disp": int(self.verbose),
            "maxcor": 250,
            "ftol": 1e-16,
            "gtol": 1e-16,
            "eps": 1e-16,
            # "eps": 1e-11,
            "maxfun": 15000,
            "maxiter": 1300,
            "iprint": 1,
            "maxls": 200,
            "finite_diff_rel_step": None,
        }

        self.options = options or default_options
        self.options["disp"] = int(self.verbose)

        self.eps_min = eps_min
        self.eps_max = eps_max

        if self.compile:
            self.fun_jit = jit(self.fun)
            self.grad_fun_jit = jit(grad(self.fun_jit))
        else:
            self.fun_jit = self.fun
            self.grad_fun_jit = grad(self.fun_jit)

    def opt2eps(self, x):
        eps_re, eps_im = x[: int(self.nvar / 2)], x[int(self.nvar / 2) :]
        return eps_re + 1j * eps_im

    def fun(self, x):
        epsilon = self.opt2eps(x) * np.ones(self.nfreq)
        self.cpw.epsilon = epsilon
        S = self.cpw.get_S(self.frequencies)
        obj = mse(self.Smeas["S11"], S["S11"])/mse(self.Smeas["S11"], 0)
        obj += mse(self.Smeas["S21"], S["S21"])/mse(self.Smeas["S21"], 0)
        obj += mse(self.Smeas["S12"], S["S12"])/mse(self.Smeas["S12"], 0)
        obj += mse(self.Smeas["S22"], S["S22"])/mse(self.Smeas["S22"], 0)
        # aa =  np.mean(np.abs(np.gradient(epsilon)/np.gradient(self.frequencies)))
        # print(npo.float64(aa))
        # obj += aa * 1e-4
        # obj += mse(self.Smeas["S12"], S["S12"])
        # obj += mse(self.Smeas["S22"], S["S22"])
        return obj

    def init_bounds(self):
        if self.bounds is not None:
            return npo.array(self.bounds)
        else:
            bounds = [
                (self.eps_min.real, self.eps_max.real)
                for i in range(int(self.nvar / 2))
            ]
            bounds += [
                (self.eps_min.imag, self.eps_max.imag)
                for i in range(int(self.nvar / 2))
            ]
            return npo.array(bounds)

    def init_guess(self):
        initial_guess = npo.complex64(self.guess * npo.ones(self.nfreq))
        return eps2opt(initial_guess)

    def run(self):
        bounds = self.init_bounds()
        initial_guess = self.init_guess()
        # print(initial_guess)
        # print(type(initial_guess))
        # print(bounds)
        # print(type(bounds))
        if self.backend == "scipy":
            opt = minimize(
                self.fun_jit,
                initial_guess,
                bounds=bounds,
                tol=1e-12,
                options=self.options,
                jac=self.jacobian,
                method="L-BFGS-B",
            )
        else:

            def fun_nlopt(x, gradn):
                y = self.fun_jit(x)
                y = npo.float64(y)
                if gradn.size > 0:
                    dy = npo.float64(self.jacobian(x))
                    gradn[:] = dy
                if self.verbose:
                    print(f">>> objective = {y}")
                    eps = self.opt2eps(x)
                    print(f"    mean permittivity = {np.mean(eps)}")
                    print(f"    std permittivity  = {np.std(eps)}")
                return y

            opt = nlopt.opt(nlopt.LD_MMA, self.nvar)
            bounds = self.init_bounds()
            opt.set_lower_bounds(bounds[:, 0])
            opt.set_upper_bounds(bounds[:, 1])
            opt.set_maxeval(self.options["maxiter"])
            opt.set_ftol_rel(self.options["ftol"])
            opt.set_xtol_rel(self.options["eps"])
            opt.set_min_objective(fun_nlopt)
            xopt = opt.optimize(initial_guess)
            fopt = opt.last_optimum_value()
            opt.x = xopt
            opt.fun = fopt

        eps_opt = self.opt2eps(opt.x) * np.ones(self.nfreq)
        self.eps_opt = eps_opt
        self.opt = opt
        
        self.cpw.epsilon = npo.array(eps_opt)
        
        return opt

    def jacobian(self, x):
        out = npo.array(self.grad_fun_jit(x))
        out = out.astype(float)
        return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["load", "get_Smeas"]

from numpy import loadtxt, interp
from .helpers import meas2complex

def load(filename):

    S = dict(
        S11=dict(A=None, phi=None),
        S21=dict(A=None, phi=None),
        S12=dict(A=None, phi=None),
        S22=dict(A=None, phi=None),
    )
    (
        f,
        S["S11"]["A"],
        S["S11"]["phi"],
        S["S21"]["A"],
        S["S21"]["phi"],
        S["S12"]["A"],
        S["S12"]["phi"],
        S["S22"]["A"],
        S["S22"]["phi"],
    ) = loadtxt(filename, skiprows=9).T

    f *= 1e-9
    return f, S



def _prepSmeas(S, frequencies, f):
    s = meas2complex(S["A"], S["phi"])
    return interp(frequencies, f, s)


def get_Smeas(filename, frequencies):
    f, S = load(filename)
    Smeas = dict()
    Smeas["S11"] = _prepSmeas(S["S11"], frequencies, f)
    Smeas["S21"] = _prepSmeas(S["S21"], frequencies, f)
    Smeas["S12"] = _prepSmeas(S["S12"], frequencies, f)
    Smeas["S22"] = _prepSmeas(S["S22"], frequencies, f)
    return Smeas
